# Standard library
from typing import (
    NamedTuple,
    Tuple,
)

Frame = NamedTuple('Frame', [
    ('event', str),
    ('function', str),
    ('level', int),
    ('timestamp', float),
])

DaemonResult = NamedTuple('DaemonResult', [
    ('stack', Tuple[Frame, ...]),
])

LoopSnapshot = NamedTuple('LoopSnapshot', [
    ('block_duration_ratio', float),
    ('real_tick_duration', float),
    ('timestamp', float),
    ('wanted_tick_duration', float),
])



# Standard library
import asyncio
import contextlib
import contextvars
import functools
import time
import threading
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
)

# Local libraries
from tracers.analyzers import (
    analyze_loop_snapshots,
    analyze_stack,
)
from tracers.constants import (
    LOOP_CHECK_INTERVAL,
)
from tracers.containers import (
    DaemonResult,
    Frame,
    LoopSnapshot,
)
from tracers.contextvars import (
    LEVEL,
    STACK,
    TRACING,
)
from tracers.daemon import (
    send_result_to_daemon,
)
from tracers.utils import (
    delta,
    divide,
    get_function_id,
    increase_counter,
)

# Types
FunctionWrapper = Callable[[Callable[..., Any]], Callable[..., Any]]


def measure_loop_skew(
    clock_id: int,
    should_measure_loop_skew: threading.Event,
    snapshots: List[LoopSnapshot],
) -> None:

    def callback_handler(
        wanted_tick_duration: float,
        start_timestamp: float,
    ) -> None:
        real_tick_duration: float = \
            delta(start_timestamp, time.clock_gettime(clock_id))

        snapshots.append(LoopSnapshot(
            block_duration_ratio=divide(
                numerator=real_tick_duration,
                denominator=wanted_tick_duration,
                on_zero_denominator=1.0,
            ),
            real_tick_duration=real_tick_duration,
            timestamp=start_timestamp,
            wanted_tick_duration=wanted_tick_duration,
        ))

        schedule_callback(wanted_tick_duration)

    def schedule_callback(wanted_tick_duration: float) -> None:
        if should_measure_loop_skew.is_set():
            with contextlib.suppress(RuntimeError):
                callback_handler_args: Tuple[float, float] = (
                    wanted_tick_duration, time.clock_gettime(clock_id),
                )

                asyncio.get_running_loop().call_later(
                    wanted_tick_duration,
                    callback_handler,
                    *callback_handler_args,
                )

    schedule_callback(LOOP_CHECK_INTERVAL)


def record(
    clock_id: int,
    event: str,
    function: Callable[..., Any],
    function_name: str = '',
) -> None:
    STACK.set(STACK.get() + (Frame(
        event=event,
        function=function_name or get_function_id(function),
        level=LEVEL.get(),
        timestamp=time.clock_gettime(clock_id),
    ),))


def _get_wrapper(  # noqa: MC001
    *,
    clock_id: int = time.CLOCK_MONOTONIC,
    do_trace: bool = True,
    function_name: str = '',
) -> FunctionWrapper:

    def decorator(function: Callable[..., Any]) -> Callable[..., Any]:

        if asyncio.iscoroutinefunction(function):

            @functools.wraps(function)
            async def isolated_wrapper(*args: Any, **kwargs: Any) -> Any:
                if do_trace and TRACING.get():
                    should_measure_loop_skew = threading.Event()

                    with increase_counter(LEVEL):
                        if LEVEL.get() == 1:
                            should_measure_loop_skew.set()
                            snapshots: List[LoopSnapshot] = []
                            measure_loop_skew(
                                clock_id,
                                should_measure_loop_skew,
                                snapshots,
                            )

                        record(clock_id, 'call', function, function_name)
                        result = await function(*args, **kwargs)
                        record(clock_id, 'return', function, function_name)

                        if LEVEL.get() == 1:
                            stack = STACK.get()
                            analyze_stack(stack)
                            analyze_loop_snapshots(tuple(snapshots))
                            send_result_to_daemon(
                                result=DaemonResult(
                                    stack=stack,
                                ),
                            )
                            should_measure_loop_skew.clear()
                else:
                    # Disable downstream tracers
                    TRACING.set(False)

                    # No overhead is introduced!
                    result = await function(*args, **kwargs)

                return result

        elif callable(function):

            @functools.wraps(function)
            def isolated_wrapper(*args: Any, **kwargs: Any) -> Any:

                @functools.wraps(function)
                def wrapper(*args: Any, **kwargs: Any) -> Any:
                    if do_trace and TRACING.get():
                        with increase_counter(LEVEL):
                            record(clock_id, 'call', function, function_name)
                            result = function(*args, **kwargs)
                            record(clock_id, 'return', function, function_name)

                            if LEVEL.get() == 1:
                                stack = STACK.get()
                                analyze_stack(stack)
                                send_result_to_daemon(
                                    result=DaemonResult(
                                        stack=stack,
                                    ),
                                )
                    else:
                        # Disable downstream tracers
                        TRACING.set(False)

                        # No overhead is introduced!
                        result = function(*args, **kwargs)

                    return result

                if LEVEL.get() == 0:
                    result = contextvars.copy_context().run(
                        wrapper, *args, **kwargs
                    )
                else:
                    result = wrapper(*args, **kwargs)

                return result

        else:

            # We were not able to wrap this object
            raise TypeError(
                f'Excpected callable or coroutine function, '
                f'got: {type(function)}'
            )

        return isolated_wrapper

    return decorator


def trace_process(
    *,
    do_trace: bool = True,
    function_name: str = '',
) -> FunctionWrapper:
    return _get_wrapper(
        clock_id=time.CLOCK_PROCESS_CPUTIME_ID,
        do_trace=do_trace,
        function_name=function_name,
    )


def trace_thread(
    *,
    do_trace: bool = True,
    function_name: str = '',
) -> FunctionWrapper:
    return _get_wrapper(
        clock_id=time.CLOCK_THREAD_CPUTIME_ID,
        do_trace=do_trace,
        function_name=function_name,
    )


def trace_monotonic(
    *,
    do_trace: bool = True,
    function_name: str = '',
) -> FunctionWrapper:
    return _get_wrapper(
        clock_id=time.CLOCK_MONOTONIC,
        do_trace=do_trace,
        function_name=function_name,
    )


def trace(
    function: Optional[Callable[..., Any]] = None,
    *,
    do_trace: bool = True,
    function_name: str = '',
) -> FunctionWrapper:
    wrapper: FunctionWrapper = trace_monotonic(
        do_trace=do_trace,
        function_name=function_name,
    )

    if function:
        return wrapper(function)

    return wrapper



# Standard library
from typing import (
    Optional,
)

# Third party libraries
from aiogqlc import GraphQLClient

# Local libraries
from tracers.config import (
    CONFIG,
)

if all([
    CONFIG.api_token,
    CONFIG.endpoint_url,
]):
    CLIENT: Optional[GraphQLClient] = GraphQLClient(
        endpoint=CONFIG.endpoint_url,
        headers={
            'authorization': f'Bearer {CONFIG.api_token}'
        },
    )
else:
    CLIENT = None



# Standard library
import asyncio
import contextlib
from contextvars import (
    ContextVar,
    Token,
)
import sys
from typing import (
    Any,
    Callable,
    Iterator,
)


def delta(start_timestamp: float, end_timestamp: float) -> float:
    return end_timestamp - start_timestamp


def divide(
    *,
    numerator: float,
    denominator: float,
    on_zero_denominator: float,
) -> float:
    return \
        on_zero_denominator if denominator == 0.0 else numerator / denominator


def get_function_id(
    function: Callable[..., Any],
) -> str:
    # Adding decorators to a function modify its metadata
    #   Fortunately functools' wrapped functions keep a reference to the parent
    while hasattr(function, '__wrapped__'):
        function = getattr(function, '__wrapped__')

    module: str = function.__module__
    name: str = function.__name__
    prefix = 'async ' * asyncio.iscoroutinefunction(function)

    if module not in {'__main__'}:
        return f'{prefix}{module}.{name}'

    return f'{prefix}{name}'


@contextlib.contextmanager
def increase_counter(contextvar: ContextVar[int]) -> Iterator[None]:
    token: Token[int] = contextvar.set(contextvar.get() + 1)
    try:
        yield
    finally:
        contextvar.reset(token)


def log(*parts: Any) -> None:
    print(*parts)


def log_stderr(*parts: Any) -> None:
    print(*parts, file=sys.stderr)



# Standard library
import os
from typing import (
    NamedTuple,
    Optional,
)

# Local libraries
from tracers.utils import (
    log_stderr,
)


def _get(var_name: str) -> Optional[str]:
    var_value: Optional[str] = os.environ.get(var_name)

    if var_name in [
        'TRACERS_API_TOKEN',
    ]:
        log_stderr(f'[INFO] {var_name} = {"<set>" if var_value else ""}')
    else:
        log_stderr(f'[INFO] {var_name} = {var_value}')

    return var_value


# Runtime constants
Config = NamedTuple('Config', [
    ('api_token', Optional[str]),
    ('endpoint_url', Optional[str]),
])

CONFIG = Config(
    api_token=_get('TRACERS_API_TOKEN'),
    endpoint_url=_get('TRACERS_ENDPOINT_URL'),
)



# Standard library
import asyncio
from collections import deque
import json
from threading import Thread
from typing import (
    Deque,
    Tuple,
)

# Third party libraries
import aiohttp
import aiogqlc
from more_itertools import iter_except

# Local libraries
from tracers.containers import (
    DaemonResult,
)
from tracers.graphql import (
    CLIENT as GRAPHQL_CLIENT,
)
from tracers.utils import (
    delta,
    log_stderr,
)

# Private constants
_RESULTS_QUEUE: Deque[DaemonResult] = deque()


async def daemon() -> None:
    while await asyncio.sleep(1.0, result=True):
        results = tuple(iter_except(_RESULTS_QUEUE.pop, IndexError))
        results_len = len(results)

        if GRAPHQL_CLIENT and results:
            success, msg = await send_results_to_server(
                client=GRAPHQL_CLIENT,
                results=results,
            )

            if success:
                log_stderr(f'[DAEMON] Uploaded transactions: {results_len}')
            else:
                log_stderr(f'[DAEMON] Error uploading: {msg}')


def send_result_to_daemon(
    *,
    result: DaemonResult,
) -> None:
    _RESULTS_QUEUE.appendleft(result)


async def send_results_to_server(
    *,
    client: aiogqlc.GraphQLClient,
    results: Tuple[DaemonResult, ...],
) -> Tuple[bool, str]:
    try:
        await client.execute(
            query="""
                mutation(
                    $transactions: [TransactionInput!]!
                ) {
                    putTransaction(
                        transactions: $transactions
                    ) {
                        success
                    }
                }
            """,
            variables=dict(
                transactions=[
                    dict(
                        initiator=result.stack[0].function,
                        stack=json.dumps(result.stack),
                        tenantId='123',
                        totalTime=str(delta(
                            result.stack[0].timestamp,
                            result.stack[-1].timestamp,
                        )),
                    )
                    for result in results
                ],
            ),
        )
    except aiohttp.ClientError as exception:
        msg: str = str(exception)
        success: bool = False
    else:
        msg = ''
        success = True

    return success, msg


# Side effect: Start an asynchronous daemon server
Thread(
    daemon=True,
    name='Tracers Daemon',
    target=lambda: asyncio.run(daemon()),
).start()



# Standard library
from contextvars import (
    ContextVar,
)
from typing import (
    Tuple,
)

# Local libraries
from tracers.containers import (
    Frame,
)

LEVEL: ContextVar[int] = \
    ContextVar('LEVEL', default=0)
STACK: ContextVar[Tuple[Frame, ...]] = \
    ContextVar('STACK', default=())
TRACING: ContextVar[bool] = \
    ContextVar('TRACING', default=True)



# Standard library
from itertools import groupby
from operator import (
    attrgetter,
    itemgetter,
)
from typing import (
    NamedTuple,
    Tuple,
)

# Local libraries
from tracers.containers import (
    Frame,
    LoopSnapshot,
)
from tracers.constants import (
    CHAR_BROKEN_BAR,
    CHAR_CHECK_MARK,
    CHAR_INFO,
    CHAR_SPACE,
    CHAR_SUPERSCRIPT_ONE,
    LOOP_SKEW_TOLERANCE,
)
from tracers.utils import (
    delta,
    divide,
    log,
)

Result = NamedTuple('Result', [
    ('counter', int),
    ('function', str),
    ('indentation', str),
    ('level', int),
    ('net_time_ratio', float),
    ('net_time_seconds', float),
    ('raw_time_ratio', float),
    ('raw_time_seconds', float),
    ('relative_timestamp', float),
])


def analyze_loop_snapshots(
    snapshots: Tuple[LoopSnapshot, ...],
) -> None:
    top_snapshots: Tuple[LoopSnapshot, ...] = tuple(sorted(
        [
            snapshot
            for snapshot in snapshots
            if snapshot.block_duration_ratio > 1.0 + LOOP_SKEW_TOLERANCE
        ],
        key=attrgetter('real_tick_duration'),
        reverse=True,
    ))

    if top_snapshots:
        log()
        log('  Some blocks (skews) occurred in the event loop',
            CHAR_SUPERSCRIPT_ONE)
        log()
        log('  #    Timestamp     Delay')
        log()
        initial_timestamp: float = snapshots[0].timestamp
        for counter, snapshot in enumerate(top_snapshots):
            skew: float = delta(
                snapshot.wanted_tick_duration,
                snapshot.real_tick_duration,
            )

            log(f'{counter:>6}',
                f'{delta(initial_timestamp, snapshot.timestamp):>8.2f}s',
                f'{skew:>8.2f}s')
        log()
        log(f'  {CHAR_SUPERSCRIPT_ONE}',
            'Consider reviewing them carefully',
            'to improve the overall system throughput')


def analyze_stack(  # pylint: disable=too-many-locals
    stack: Tuple[Frame, ...],
) -> None:
    stack_levels: Tuple[int, ...] = \
        tuple(map(attrgetter('level'), stack))

    total_time_seconds: float = \
        delta(stack[0].timestamp, stack[-1].timestamp)

    log()
    log(f'{CHAR_INFO} Finished transaction: {total_time_seconds:.2f} seconds')
    log()
    log('     # Timestamp                Net              Total    Call Chain')
    log()

    counter: int = 0
    results: Tuple[Result, ...] = ()
    accumulator: Tuple[Result, ...] = ()
    for index, frame in enumerate(stack):
        if frame.event == 'call':
            counter += 1

            frame_childs: Tuple[Frame, ...] = \
                stack[index:stack_levels.index(frame.level, index + 1) + 1]

            raw_time_seconds: float = \
                delta(frame_childs[0].timestamp, frame_childs[-1].timestamp)

            net_time_seconds: float = \
                raw_time_seconds - sum([
                    +x.timestamp if x.event == 'return' else -x.timestamp
                    for x in frame_childs
                    if x.level == frame.level + 1
                ])

            results += (Result(
                counter=counter,
                function=frame.function,
                indentation=(
                    (3 * CHAR_SPACE + CHAR_BROKEN_BAR) * (frame.level - 1) +
                    (3 * CHAR_SPACE + CHAR_CHECK_MARK)
                ),
                level=frame.level,
                net_time_ratio=100.0 * divide(
                    numerator=net_time_seconds,
                    denominator=total_time_seconds,
                    on_zero_denominator=1.0,
                ),
                net_time_seconds=net_time_seconds,
                relative_timestamp=delta(stack[0].timestamp, frame.timestamp),
                raw_time_ratio=100.0 * divide(
                    numerator=raw_time_seconds,
                    denominator=total_time_seconds,
                    on_zero_denominator=1.0,
                ),
                raw_time_seconds=raw_time_seconds,
            ),)

    if results:
        accumulator += (results[0],)
        accumulator = flush_accumulator(accumulator)

    for index, result in enumerate(results[1:-1], start=1):
        accumulator += (result,)

        if ((result.level != results[index - 1].level or
                result.function != results[index - 1].function) and
            (result.level != results[index + 1].level or
                result.function != results[index + 1].function)):
            accumulator = flush_accumulator(accumulator)

    if len(results) > 1:
        accumulator += (results[-1],)
        accumulator = flush_accumulator(accumulator)

    log()
    log('           Count                Net              Total    Function')
    log()

    key = attrgetter('function')

    for (
        function,
        net_time_ratio,
        net_time_seconds,
        raw_time_ratio,
        raw_time_seconds,
        times_called
    ) in sorted(
        (
            (
                function,
                sum(map(attrgetter('net_time_ratio'), group)),
                sum(map(attrgetter('net_time_seconds'), group)),
                sum(map(attrgetter('raw_time_ratio'), group)),
                sum(map(attrgetter('raw_time_seconds'), group)),
                len(group),
            )
            for function, group_iter in groupby(
                sorted(
                    results,
                    key=key,
                ),
                key=key,
            )
            for group in [tuple(group_iter)]
        ),
        key=itemgetter(2),
        reverse=True,
    ):
        log(
            f'{times_called:>16}',
            f'{net_time_seconds:>8.2f}s',
            f'[{net_time_ratio:>5.1f}%]',
            f'{raw_time_seconds:>8.2f}s',
            f'[{raw_time_ratio:>5.1f}%]',
            f'{3 * CHAR_SPACE + CHAR_CHECK_MARK}',
            f'{function}',
        )


def flush_accumulator(accumulator: Tuple[Result, ...]) -> Tuple[Result, ...]:
    if accumulator:
        times: str = f'{len(accumulator)} times: ' * (len(accumulator) > 1)

        log(
            f'{accumulator[0].counter:>6}',
            f'{accumulator[0].relative_timestamp:>8.2f}s',
            f'{sum(map(attrgetter("net_time_seconds"), accumulator)):>8.2f}s',
            f'[{sum(map(attrgetter("net_time_ratio"), accumulator)):>5.1f}%]',
            f'{sum(map(attrgetter("raw_time_seconds"), accumulator)):>8.2f}s',
            f'[{sum(map(attrgetter("raw_time_ratio"), accumulator)):>5.1f}%]',
            f'{accumulator[0].indentation}',
            f'{times}{accumulator[0].function}',
        )

    return ()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tracers', 'tracers.function']

package_data = \
{'': ['*']}

install_requires = \
['aioboto3==8.0.3', 'aiogqlc==1.0.5', 'more-itertools==8.4.0']

setup_kwargs = {
    'name': 'tracers',
    'version': '20.6.39034',
    'description': 'Open-Source APM (Application monitoring) project that offers you zero overhead wrappers for profiling your code execution flow',
    'long_description': "# Table of contents\n\n- [Table of contents](#table-of-contents)\n- [Project description](#project-description)\n  * [Key Features](#key-features)\n- [Quick Introduction](#quick-introduction)\n- [Installation](#installation)\n- [Examples](#examples)\n\n# Project description\n\n**Tracers** is an Open-Source **APM** (Application monitoring) project\nthat offers you zero overhead wrappers for profiling your code execution flow\n\n```\n🛈  Finished transaction: 181019281c674f53b3bdbf5a0e4cac35, 3.81 seconds\n\n     # Timestamp                Net              Total    Call Chain\n\n     1     0.00s     0.00s [  0.0%]     3.81s [100.0%]    ✓ async function_a\n     2     0.00s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ✓ async asyncio.tasks.sleep\n     3     0.10s     0.50s [ 13.1%]     0.50s [ 13.1%]    ¦   ✓ time.sleep\n     4     0.60s     0.00s [  0.0%]     3.21s [ 84.2%]    ¦   ✓ async function_b\n     5     0.60s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ✓ async asyncio.tasks.sleep\n     6     0.70s     0.00s [  0.0%]     0.70s [ 18.4%]    ¦   ¦   ✓ async function_c\n     7     0.70s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ¦   ✓ async asyncio.tasks.sleep\n     8     0.80s     0.50s [ 13.1%]     0.50s [ 13.1%]    ¦   ¦   ¦   ✓ time.sleep\n     9     1.30s     0.00s [  0.0%]     0.10s [  2.6%]    ¦   ¦   ¦   ✓ async function_d\n    10     1.30s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ¦   ¦   ✓ async asyncio.tasks.sleep\n    11     1.41s     2.00s [ 52.5%]     2.00s [ 52.5%]    ¦   ¦   ✓ time.sleep\n    12     3.41s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ✓ async asyncio.tasks.sleep\n    13     3.51s     0.00s [  0.0%]     0.10s [  2.6%]    ¦   ¦   ✓ async function_d\n    14     3.51s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ¦   ✓ async asyncio.tasks.sleep\n    15     3.61s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ✓ async asyncio.tasks.sleep\n    16     3.71s     0.00s [  0.0%]     0.10s [  2.6%]    ¦   ¦   ✓ async function_e\n    17     3.71s     0.10s [  2.6%]     0.10s [  2.6%]    ¦   ¦   ¦   ✓ async asyncio.tasks.sleep\n\n           Count                Net              Total    Function\n\n               3     3.00s [ 78.8%]     3.00s [ 78.8%]    ✓ time.sleep\n               8     0.81s [ 21.1%]     0.81s [ 21.1%]    ✓ async asyncio.tasks.sleep\n               1     0.00s [  0.0%]     3.21s [ 84.2%]    ✓ async function_b\n               1     0.00s [  0.0%]     0.70s [ 18.4%]    ✓ async function_c\n               1     0.00s [  0.0%]     3.81s [100.0%]    ✓ async function_a\n               2     0.00s [  0.0%]     0.20s [  5.3%]    ✓ async function_d\n               1     0.00s [  0.0%]     0.10s [  2.6%]    ✓ async function_e\n\n\n  Some blocks (skews) occurred in the event loop ¹\n\n  #    Timestamp     Delay\n\n     0     1.40s     2.00s\n     1     0.09s     0.50s\n     2     0.80s     0.50s\n\n  ¹ Consider reviewing them carefully to improve the overall system throughput\n```\n\n## Key Features\n\n- Handles **any callable** object, which includes **your own code**,\n  **third party libraries**, and even the low-level **Python standard library**\n- Handles [**async**](https://docs.python.org/3/library/asyncio.html) code\n  **out-of-the box**, no config required\n- Exposes a **high-level API**:\n  - `@trace` decorator (which internally handles async/sync cases)\n- It's **Thread-safe**, **Async-safe**, **Process-safe** and **Context-safe**\n  - Accurate results in any scenario\n  - No worries about leaking, bleeding, corrupting, or locking stuff into other\n    code\n- Introduces **zero overhead** in production!\n  - The `@trace` decorator accepts a `do_trace` parameter\n    that you can dynamically set to `True` of `False` to differentiate\n    testing environments from production environments\n- It's **easy to deploy**\n  - No external dependencies!\n- It's easy to pin-point performance problems:\n  - Gives you the total execution time in seconds and **%**\n  - Allows you to identify points in time where your **async** event loop got blocked\n- Allows you to measure monotonic (wall time), process time, and thread time out-of-the box\n- Profiles without using dirty introspection stuff\n  - The main code is just 50 lines long, pretty high level, go and read it :)\n\n# Quick Introduction\n\nLet's start with a very basic example:\n\n```py\nimport time\nfrom dateutil.parser import parse\n\n\ndef example():\n    time.sleep(2.0)\n    your_business_logic('Sat Oct 11')\n\n\ndef your_business_logic(date: str):\n    parse(date)\n    time.sleep(1.0)\n\n\nexample()\n```\n\nTracing its flow and gathering profiling information is a matter of\ndecorating your functions:\n\n```diff\n--- a/examples/without_tracers.py\n+++ b/examples/with_tracers.py\n@@ -1,15 +1,18 @@\n import time\n from dateutil.parser import parse\n+from tracers.function import trace\n\n\n+@trace\n def example():\n     time.sleep(2.0)\n     your_business_logic('Sat Oct 11')\n\n\n+@trace\n def your_business_logic(date: str):\n     parse(date)\n     time.sleep(1.0)\n\n\n example()\n```\n\nIf you run it, all the functions you decorated will be traced\nand you'll have metrics of the execution flow:\n\n```\n🛈  Finished transaction: d5fa597d99904fc0b5f06f2b66d8b0ad, 3.00 seconds\n\n     # Timestamp                Net              Total    Call Chain\n\n     1     0.00s     2.00s [ 66.7%]     3.00s [100.0%]    ✓ example\n     2     2.00s     1.00s [ 33.3%]     1.00s [ 33.3%]    ¦   ✓ your_business_logic\n```\n\nFrom the output you can conclude:\n- executing function *example* took a total of *3.0* seconds to complete\n- function *example* takes *66.7%* of the execution time\n- function *example encloses *100%* of your code time\n- function *example* called function: *your_business_logic*\n- function *your_business_logic* took *1.0* seconds out of the *3.0* seconds\n  the function *example* needed to complete\n- function *your_business_logic* represents *33.3%* of your execution time\n\nTracing code is not limited to your own code.\nYou can trace any **callable object** including **third party packages**,\n**Python's standard library**, and almost anything\n\nThe level of detail is up to you!\n\n\n```diff\n--- a/examples/with_tracers.py\n+++ b/examples/with_detailed_tracers.py\n@@ -1,18 +1,18 @@\n import time\n from dateutil.parser import parse\n from tracers.function import trace\n\n\n @trace\n def example():\n-    time.sleep(2.0)\n+    trace(time.sleep)(2.0)\n     your_business_logic('Sat Oct 11')\n\n\n @trace\n def your_business_logic(date: str):\n-    parse(date)\n-    time.sleep(1.0)\n+    trace(parse)(date)\n+    trace(time.sleep)(1.0)\n\n\n example()\n```\n\n```\n🛈  Finished transaction: b81e4bfbc785458994924b87c6baa668, 3.00 seconds\n\n     # Timestamp                Net              Total    Call Chain\n\n     1     0.00s     0.00s [  0.0%]     3.00s [100.0%]    ✓ example\n     2     0.00s     2.00s [ 66.6%]     2.00s [ 66.6%]    ¦   ✓ time.sleep\n     3     2.00s     0.00s [  0.0%]     1.00s [ 33.4%]    ¦   ✓ your_business_logic\n     4     2.00s     0.00s [  0.0%]     0.00s [  0.0%]    ¦   ¦   ✓ dateutil.parser._parser.parse\n     5     2.00s     1.00s [ 33.3%]     1.00s [ 33.3%]    ¦   ¦   ✓ time.sleep\n```\n\n# Installation\n\nWe are hosted on **PyPI**: https://pypi.org/project/tracers\n\nJust run: `pip install tracers`\nor use the package manager you like the most\n\n# Examples\n\nCheck them out in the\n[examples](https://github.com/kamadorueda/tracers/tree/master/examples)\nfolder\n",
    'author': 'Kevin Amado',
    'author_email': 'kamadorueda@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/kamadorueda/tracers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
